// SPHGEN_CPP Sphere.cpp
//
// Andrew Magis <amagis@cise.ufl.edu>
// Last change: Time-stamp: <26-Jul-2006 10:22:41 amagis>
//
// Version 1.1.  
// This software is open source, to be used in conjunction with the UCSF DOCK
// package.  If you have questions, comments, or suggestions, please contact
// me at the email address above.
//

#include "Sphere.hpp"

//Default constructor
Vector3d::Vector3d()
	:x(0.0), y(0.0), z(0.0)
{
	c[0] = 0.0; c[1] = 0.0; c[2] = 0.0;
}

Vector3d::Vector3d(double _x, double _y, double _z)
	:x(_x), y(_y), z(_z)
{
	c[0] = _x; c[1] = _y; c[2] = _z;
}

Vector3d::~Vector3d()
{}

//Returns the sum of the two vectors
Vector3d Vector3d::operator +(const Vector3d &v)
{
	return Vector3d(x+v.x, y+v.y, z+v.z);
}

//Returns the difference of the two vectors
Vector3d Vector3d::operator -(const Vector3d &v)
{
	return Vector3d(x-v.x, y-v.y, z-v.z);
}

//Returns assignment of vector
Vector3d& Vector3d::operator =(const Vector3d &v)
{
	x = v.x, y = v.y, z = v.z;
	return *this;
}

//Returns the dot product of the two vectors
double Vector3d::operator *(const Vector3d &v)
{
	return (x*v.x + y*v.y + z*v.z);
}

//Prints the vector
void Vector3d::Print(FILE *out)
{
	fprintf(out, "%lf %lf %lf", x, y, z);
}


//Default Constructor
Atom::Atom()
	:num_surfacepoints(0), atom_number(-1),counter(0), num_spheres(0),
	 points(NULL), position(NULL), largest(NULL), rtmax(0.0)
{}

Atom::Atom(double _x, double _y, double _z, char *_residue_name,
		   char *_residue, char *_atom_name, int _num_surfacepoints,
		   int _atom_number)
	:num_surfacepoints(_num_surfacepoints), atom_number(_atom_number),
	 counter(0), num_spheres(0), largest(NULL), rtmax(0.0)
{

	//Allocate the surface point array
	points = new SurfacePoint*[num_surfacepoints];
	
	//Allocate the position of this atom
	position = new Vector3d(_x, _y, _z);

	//Copy in label information.
	strncpy(residue_name, _residue_name, 3);
	strncpy(atom_name, _atom_name, 3);
	strncpy(residue, _residue, 4);
}

//Destructor
Atom::~Atom()
{
	for (int i = 0; i < num_surfacepoints; i++)
		if (points[i] != NULL)
			delete points[i];

	for (int i = 0; i < num_spheres; i++)
		if (spheres[i] != NULL) {
			if (spheres[i]->cs != NULL)
				delete spheres[i]->cs;
			if (spheres[i]->pairs != NULL)
				delete[] spheres[i]->pairs;
			delete spheres[i];
		}

	if (spheres != NULL)
		delete[] spheres;
	if (points != NULL)
		delete[] points;
	if (position != NULL)
		delete position;
}

bool Atom::AddSurfacePoint(char *_tag, double _x, double _y, double _z,
						   double _nx, double _ny, double _nz)
{
	//Check to make sure surface normal is not all zeros
	if ((_nx == 0.0) && (_ny == 0.0) && (_nz == 0.0)) {
		fprintf(stderr, "ERROR! Surface normal is all zeros\n");
		return false;
	}
	
	//Allocate the new surface point	
	if (counter < num_surfacepoints) {
		points[counter++] = new SurfacePoint(_tag, _x, _y, _z,_nx,
											 _ny, _nz, atom_number);
		return true;
	}
	return false;
}

int Atom::CompareAtoms(Atom *a)
{

	//For each surface point of this atom, compare with all other
	//surface points of Atom a, returning the MAX number of spheres
	//possible for this atom
	for (int i = 0; i < num_surfacepoints; i++) {
		num_spheres += points[i]->CompareSurfacePoints(a);

	}
	return 1;	
}

//Returns true if points are close enough
int Atom::GenerateSpheres(double dotlim, double radmin, double radmax)
{

	//Allocate the sphere array for this atom, using the maximum number
	//of spheres possible.  Some may not be actually made
	spheres = new Sphere*[num_spheres];
	num_spheres = 0;
	
	//Actually generate the spheres
	for (int i = 0; i < num_surfacepoints; i++) {

		//Call each surface point
		spheres[num_spheres] = points[i]->WithinLimits(dotlim, radmin);
		if (spheres[num_spheres] != NULL) {
			
			//Keep track of the largest sphere for this atom
			if (spheres[num_spheres]->rtmin > rtmax) {
				rtmax = spheres[num_spheres]->rtmin;
				largest = spheres[num_spheres];
			}
			//Count actual number of spheres created
			num_spheres++;
		}
	}

	//If the largest sphere generated is smaller than the maximum radius
	if ((largest != NULL) && (largest->rtmin < radmax))
		return 1;
	
	largest = NULL;
	return 0;
}

//Prints the atom information and the surface points
void Atom::Print(FILE *out)
{
	for (int i = 0; i < counter; i++) {
		fprintf(out, "%.3s %.3s %.3s ", residue_name, residue, atom_name);
		points[i]->Print(out);
		printf("\n");
	}
}

//Default constructor
SurfacePoint::SurfacePoint()
	:position(NULL), normal(NULL), nearest(NULL)
{}

//Constructor
SurfacePoint::SurfacePoint(char *_tag, double _x, double _y, double _z,
						   double _nx, double _ny, double _nz, int _atom_number)
	:atom_number(_atom_number), nearest(NULL), rmin(1000000.0),
	 rtmin(1000.0), rtmina(3500), rtminb(2830), rtminc(2000)
{
	strncpy(tag, _tag, 3);
	position = new Vector3d(_x, _y, _z);
	normal = new Vector3d(_nx, _ny, _nz);

	//Calculate the dot product of the position with itself
	dotproduct = *position * *position;

	//Determine if any of the normal coordinates are zero
	izero = 0;
	if (normal->x == 0.0) izero++;
	if (normal->y == 0.0) izero++;
	if (normal->z == 0.0) izero++;
}

//Destructor
SurfacePoint::~SurfacePoint()
{
	if (position != NULL)
		delete position;
	if (normal != NULL)
		delete normal;
}

//This function takes a surface point and compares it with all the sp's of an atom
int SurfacePoint::CompareSurfacePoints(Atom *a)
{
	
	double xp, yp, zp, denom, s, dotck, del[3];
	double csq, cp, slop, bet, cpa, cpb, dis;
	int js = 0, ja, jb;
	Vector3d *v;
	int max_spheres = 0;

	switch(izero) {
		
	case 0:

		//Calculate slope and beta
		slope[0] = normal->y / normal->x;
		slope[1] = normal->z / normal->x;
		beta[0] = position->y - slope[0]*position->x;
		beta[1] = position->z - slope[1]*position->x;
		
		for (int i = 0; i < a->GetNumSurfacePoints(); i++) {

			v = a->GetSurfacePoint(i)->GetPosition();

			//Calculate del (used in all three cases)
			del[0] = v->x - position->x;
			del[1] = v->y - position->y;
			del[2] = v->z - position->z;
			
			dotck = del[0]*normal->x + del[1]*normal->y + del[2]*normal->z;
			if (dotck < 0) continue;
			
			s = fabs(del[0]) + fabs(del[1]) + fabs(del[2]);
			if (s > rtmina) continue;
			
			denom = 2.0*(del[0] + del[1]*slope[0] + del[2]*slope[1]);
			if (denom == 0.0) continue;
			
			xp = (a->GetSurfacePoint(i)->GetDotProduct()-dotproduct-2.0
				  * (del[1]*beta[0] + del[2]*beta[1])) / denom;
			yp = slope[0]*xp + beta[0];
			zp = slope[1]*xp + beta[1];
			
			dis = (xp-position->x)*(xp-position->x)
				+ (yp-position->y)*(yp-position->y)
				+ (zp-position->z)*(zp-position->z);
			
			//If the distance between the sphere is less than the min radius
			if (dis < rmin) {
				nearest = a->GetSurfacePoint(i);
				rmin = dis;
				rtmin = sqrt(dis);
				rtmina = 3.5*rtmin;
				cs[0] = xp; cs[1] = yp; cs[2] = zp;
				max_spheres++;
			}
		}
		break;

	case 2:

		cs[0] = position->x; cs[1] = position->y; cs[2] = position->z;
		if (fabs(normal->x) > 0.99) js = 0;
		if (fabs(normal->y) > 0.99) js = 1;
		if (fabs(normal->z) > 0.99) js = 2;
		
		switch(js) {
		case 0:
			ja = 1; jb = 2; break;
		case 1:
			ja = 0; jb = 2; break;
		default:
			ja = 0; jb = 1; break;
		}
		csq = position->c[js]*position->c[js];
			
		for (int i = 0; i < a->GetNumSurfacePoints(); i++) {

			v = a->GetSurfacePoint(i)->GetPosition();

			//Calculate del (used in all three cases)
			del[0] = v->x - position->x;
			del[1] = v->y - position->y;
			del[2] = v->z - position->z;

			//Symmetry related points can fail here if del[js] == 0
			if (del[js] == 0.0) continue;
			if (fabs(del[js]) > rtminc) continue;
			
			dotck = del[0]*normal->x + del[1]*normal->y + del[2]*normal->z;
			if (dotck < 0) continue;

			cp = ( v->c[js]*v->c[js] - csq
				   + del[ja]*del[ja] + del[jb]*del[jb] ) 
				/ (2.0 * del[js]);
			dis = (cp - position->c[js])*(cp - position->c[js]);
			
			if (dis < rmin) {
				nearest = a->GetSurfacePoint(i);;
				rmin = dis;
				rtmin = fabs(cp - v->c[js]);
				rtminc = 2.0 * rtmin;
				cs[js] = cp;
				max_spheres++;
			}
		}
		break;
			
	case 1:

		cpa = cpb = 0;
		if (normal->x == 0.0) js = 0;
		if (normal->y == 0.0) js = 1;
		if (normal->z == 0.0) js = 2;
		
		switch(js) {
		case 0:
			ja = 1; jb = 2;
			slop = normal->z / normal->y;
			break;
		case 1:
			ja = 0; jb = 2; 
			slop = normal->z / normal->x;
			break;
		default:
			ja = 0; jb = 1;
			slop = normal->y / normal->x;
			break;
		}
		bet = position->c[jb] - slop*position->c[ja];
		csq = position->c[ja]*position->c[ja] + position->c[jb]*position->c[jb];
		
		for (int i = 0; i < a->GetNumSurfacePoints(); i++) {
			
			v = a->GetSurfacePoint(i)->GetPosition();
			
			//Calculate del (used in all three cases)
			del[0] = v->x - position->x;
			del[1] = v->y - position->y;
			del[2] = v->z - position->z;
			
			dotck = del[0]*normal->x + del[1]*normal->y + del[2]*normal->z;
			if (dotck < 0) continue;
	
			s = fabs(del[ja]) + fabs(del[jb]);
			if (s > rtminb) continue;
			
			denom = 2.0*(del[ja] + slop*del[jb]);
			if ((del[ja] == 0.0) && (del[jb] == 0.0)) {
				dis = del[js]*del[js];
				cpa = position->c[ja];
				cpb = position->c[jb];
			}
			
			//No divide-by-zero
			if (denom != 0.0) {
				cpa = (v->c[ja]*v->c[ja] + v->c[jb]*v->c[jb] -
					   csq + del[js]*del[js] - 2.0*bet*del[jb]) / denom;
				cpb = slop*cpa + bet;
				dis = (cpa-position->c[ja])*(cpa-position->c[ja])
					+ (cpb-position->c[jb])*(cpb-position->c[jb]);
			}
			
			//If the distance to this point is less than any others so far
			if (dis < rmin) {
				nearest = a->GetSurfacePoint(i);
				rmin = dis;
				rtmin = sqrt(dis);
				rtminb = 2.83*rtmin;
				cs[js] = position->c[js];
				cs[ja] = cpa;
				cs[jb] = cpb;
				max_spheres++;
			}
		}
		break;
			
		default:
		fprintf(stderr, "ERROR! Surface normal is zero\n");
		return 0;
		break;
		
	} //End of CASE statement

	//Return number of spheres that will be generated for this surface point
	return max_spheres;
}

//Returns true if points are close enough
Sphere* SurfacePoint::WithinLimits(double dotlim, double radmin)
{
	Vector3d *v;;
	double dot, dotp;
	
	if (nearest != NULL) {

		v = nearest->GetPosition();
		dot = (cs[0] - position->x)*(v->x - cs[0])
			+ (cs[1] - position->y)*(v->y - cs[1])
			+ (cs[2] - position->z)*(v->z - cs[2]);
		dotp = dot / rmin;
	
		if ((dotp > dotlim) && (rtmin < 999.0) && (rtmin > radmin)) {
							
			//Create sphere
			Sphere *s = new Sphere;
			s->i = this;
			s->this_atom = atom_number;
			s->pair_atom = nearest->GetAtomNumber();
			s->ipair = nearest;
			s->rtmin = rtmin;
			s->cs = new Vector3d(cs[0], cs[1], cs[2]);
			s->pairs = NULL;
			return s;
		}
	}
	return NULL;
}

//Prints the position of the point and the normal vector
void SurfacePoint::Print(FILE *out)
{
	position->Print(out);
	fprintf(out, " %.3s ", tag);
	normal->Print(out);
}
